#!/usr/bin/env python3

import cv2
import numpy as np
import argparse
import time
import io
import math
import sys
import statistics
from array import *

'''
Les Wright 21 June 2023
https://youtube.com/leslaboratory
A Python program to read, parse and display thermal data from the Topdon TC001 Thermal camera!
'''
print('Les Wright 21 June 2023')
print('https://youtube.com/leslaboratory')
print('A Python program to read, parse and display thermal data from the Topdon TC001 Thermal camera!')
print('')
print('Tested on Debian all features are working correctly')
print('This will work on the Pi However a number of workarounds are implemented!')
print('Seemingly there are bugs in the compiled version of cv2 that ships with the Pi!')
print('')
print('Key Bindings:')
print('')
print('a z: Increase/Decrease Blur')
print('s x: Floating High and Low Temp Label Threshold')
print('d c: Change Interpolated scale [small to fullscreen] Note: This will not change the window size on the Pi')
print('f v: Contrast')
print('g  : Toggle Fullscreen Window (note going back to windowed does not seem to work on the Pi!)')
print('r  : Toggle [r]ecording')
print('p  : Sna[p]shot')
print('m M: Cycle [for|back]wards through Color[m]aps')
print('h  : Toggle HUD')
print('u  : Toggle Celsius/Fehrenheit')
print('q  : Quit')


# Calculate maxScaleSteps based on display resolution and thermal sensor resolution
display_width=2560
display_height=1600
maxScaleSteps = min( math.trunc(display_width/256), math.trunc(display_height/192) )

print('')
print('maxScaleSteps='+str(maxScaleSteps))

#We need to know if we are running on the Pi, because openCV behaves a little oddly on all the builds!
def is_raspberrypi():
    try:
        with io.open('/sys/firmware/devicetree/base/model', 'r') as m:
            if 'raspberry pi' in m.read().lower(): return True
    except Exception: pass
    return False

isPi = is_raspberrypi()

print('isPi='+str(isPi))

parser = argparse.ArgumentParser()
parser.add_argument("--device", type=int, default=0, help="Video Device number e.g. 0, use v4l2-ctl --list-devices")
args = parser.parse_args()
	
if args.device:
	dev = args.device
else:
	dev = 0
	
#init video
cap = cv2.VideoCapture('/dev/video'+str(dev), cv2.CAP_V4L)
#cap = cv2.VideoCapture(0)
#pull in the video but do NOT automatically convert to RGB, else it breaks the temperature data!
#https://stackoverflow.com/questions/63108721/opencv-setting-videocap-property-to-cap-prop-convert-rgb-generates-weird-boolean
if isPi == True:
	cap.set(cv2.CAP_PROP_CONVERT_RGB, False)
else:
	cap.set(cv2.CAP_PROP_CONVERT_RGB, 0.0)

def toggleFullScreen():
	global dispFullscreen

	dispFullscreen = not(dispFullscreen)
	cv2.destroyWindow('Thermal') # Destroy current window and recreate resizable or fullscreen
	newWindow(dispFullscreen)

def reScale(value, resize): # reScale amoung 5 resizeable windows plus fullscreen
	global scale, width, height, newWidth, newHeight, dispFullscreen, isPi

	if value == -1 and dispFullscreen == True:
		toggleFullScreen() # Switch from fullscreen to resizeable window 
		return # Only decrement in resizeable window

	scale += value

	#maxScaleSteps = 5
	if scale > maxScaleSteps:
		scale = maxScaleSteps
		if dispFullscreen == False:
			toggleFullScreen() # maxed out resizeable window, switch to fullscreen
		return # Only increment in resizeable window
	elif scale < 1:
		scale = 1
	newWidth  = scale * width
	newHeight = scale * height

	if value != 0 and dispFullscreen == True: 
        # If attempthing to scale a FULLSCREEN, switch to resizeable window first
		toggleFullScreen()
		return

	if resize == True and dispFullscreen == False and isPi == False:
		cv2.resizeWindow('Thermal', newWidth, newHeight)

def newWindow(fullscreen):
    # Add window manager's resize border with WINDOW_GUI_EXPANDED
	if fullscreen == False:
		# This allows resizing larger and smaller
		cv2.namedWindow('Thermal', cv2.WINDOW_AUTOSIZE | cv2.WINDOW_KEEPRATIO | cv2.WINDOW_GUI_EXPANDED)
	else:
		# This allows fullscreen
		cv2.namedWindow('Thermal', cv2.WINDOW_GUI_EXPANDED)
		cv2.resizeWindow('Thermal', newWidth, newHeight)
		cv2.setWindowProperty('Thermal',cv2.WND_PROP_FULLSCREEN,cv2.WINDOW_FULLSCREEN)

	reScale(0,True)


#256x192 General settings
width  = 256 #Sensor width
height = 192 #Sensor height
scale  = 3 #default starting scale multiplier


def reAlpha(value):
	global alpha

	alpha += value
	alpha = round(alpha,1)#fix round error
	if alpha > 3.0:
		alpha = 3.0
	elif alpha < 0.0: # Should lower limit be 1.0 or 0.0 ???
		alpha = 0.0

alpha = 1.0 # Contrast control (1.0-3.0) # What should the range actually be ?

def reColormap(value):
	global colormap, cmapNames, cmaps, cmapText, cmapCurrent

	colormap    = (colormap + value) % len(cmaps) # Wrap limits preventing array bounds errors
	cmapText    = cmapNames[colormap]
	cmapCurrent = cmaps[colormap]

colormap  = 0 # Index into cmapNames[] and cmaps[]
cmapNames = ['Jet','Hot','Magma','Inferno','Plasma','Bone','Spring','Autumn','Viridis','Parula','Inv Rainbow']
cmaps     = [cv2.COLORMAP_JET,cv2.COLORMAP_HOT,cv2.COLORMAP_MAGMA,cv2.COLORMAP_INFERNO,cv2.COLORMAP_PLASMA,\
cv2.COLORMAP_BONE,cv2.COLORMAP_SPRING,cv2.COLORMAP_AUTUMN,cv2.COLORMAP_VIRIDIS,cv2.COLORMAP_PARULA,cv2.COLORMAP_RAINBOW]
reColormap(0) # Initialize cmapText and cmapCurrent based on current colormap 

# OFF_RED   = (40,40,250)
# OFF_WHITE = (200,200,200)
RED    = (0,0,255)
BLUE   = (255,0,0)
BLACK  = (0,0,0)
YELLOW = (0,255,255)
WHITE  = (255,255,255)
tempTextColor = WHITE # Yellow text over yellow text HUD doesn't work well

font=cv2.FONT_HERSHEY_SIMPLEX # Global font

dispFullscreen = False

newWindow(dispFullscreen) # Start display with a resizeable window

rad = 0 #blur radius
threshold = 2
hud = True

recording = False
rColor = WHITE if recording==False else RED

elapsed = "00:00:00"
snaptime = "None"

def CorF(celsius): # Round to 1/10th of a degree F
	global useCelsius
	return celsius if useCelsius==True else round((((celsius * 9.0)/5.0) + 32.0), 1)

useCelsius = False # Change startup Celsius/Fahrenheit display defaults here
labelCF = ' C' if useCelsius==True else ' F'

def rec():
	now = time.strftime("%Y%m%d--%H%M%S")
	#do NOT use mp4 here, it is flakey!
	videoOut = cv2.VideoWriter(now+'output.avi', cv2.VideoWriter_fourcc(*'XVID'),25, (newWidth,newHeight))
	return(videoOut)

def snapshot(heatmap):
	#I would put colons in here, but it Win throws a fit if you try and open them!
	now = time.strftime("%Y%m%d-%H%M%S") 
	snaptime = time.strftime("%H:%M:%S")
	cv2.imwrite("TC001"+now+".png", heatmap)
	return snaptime
 
def kelvin2Celsius(kelvin): # LeoDJ's Kelvin conversion algorythm, post #216
	return round( ((kelvin/64.0) - 273.15), 2 )

def drawDotLabel(row, col, label, dotColor):
	global scale, newWidth, newHeight, heatmap

	row = row * scale;
	col = col * scale;
	# WHITE halo is move visible than a black halo
	cv2.circle(heatmap, (row, col), 5, WHITE,     2) # Halo color
	cv2.circle(heatmap, (row, col), 5, dotColor, -1) # [Red|Blue] dot

	rowOffset = 10
	colOffset = 5

	# Move floating text up/down/left to keep it on screen @ display boundaries ...
	# Something strange, row/height and col/width appear switched
	# TODO - Use getTextSize() for exact label width and height
	#        w,h = cv2.getTextSize(label, font, 0.45, 2 )
	if col < (newHeight/16):
		colOffset = colOffset + 10
	elif col > (newHeight*15/16):
		colOffset = colOffset - 10

	if row > (newWidth*7/8):
		rowOffset = rowOffset - 120  

	if hud == True: # Avoid writing temps over HUD text
		if row < 166 and col < 136:
			rowOffset = rowOffset + (166 - row)

	cv2.putText(heatmap, label, (row+rowOffset, col+colOffset),\
	font, 0.45,  BLACK, 2, cv2.LINE_AA)
	cv2.putText(heatmap, label, (row+rowOffset, col+colOffset),\
	font, 0.45, tempTextColor, 1, cv2.LINE_AA)


frameCounter = 0
nsA = time.time_ns()

while(cap.isOpened()):

	# Capture frame-by-frame
	ret, frame = cap.read()
	if ret == False:
		print('Capture error')

	if ret == True:
		imdata,thdata = np.array_split(frame, 2)

		#now parse the data from the bottom frame and convert to temp!

		#https://www.eevblog.com/forum/thermal-imaging/infiray-and-their-p2-pro-discussion/200/
		#Huge props to LeoDJ for figuring out how the data is stored and how to compute temp from it.
		#grab data from the center pixel...

#width  = 256 #Sensor width
#height = 192 #Sensor height
# MIN/MAX?AVG BUG: Min/Max/Avg was only calculated on the low byte, not low and high byte


		kmin = sys.maxsize 
		kmax = -(sys.maxsize - 1)
		posmin = 0
		posmax = 0
		ktotal = 0
		index = -1 
		for row in range( 0, 192 ):     # for (h = 0, h < 192, h++) height
			for col in range( 0, 256 ): # for (w = 0; w < 256, w++) width
				index = index + 1

# print(' ', thdata[row][col][0], (thdata[row][col][1] * 256) # LSByte + MSByte


				kelvin = thdata[row][col][0] + (thdata[row][col][1] * 256) # LSByte + MSByte

				ktotal = ktotal + kelvin

				if (kmin > kelvin):
					kmin = kelvin
					posmin = index

				if (kmax < kelvin):
					kmax = kelvin
					posmax = index

				#print( '[', thdata[row][col][0], ', ', thdata[row][col][1], '] ', end='', sep='')
			#sys.exit()

		# Grab crosshair outside of the loop
		MAX_TEMPS = (192 * 256) # height * width
		kch = thdata[96][128][0] + (thdata[96][128][1] * 256) # LSByte + MSByte
		kavg = ktotal / MAX_TEMPS

		mcol,mrow = divmod(posmax,width)
		lcol,lrow = divmod(posmin,width)

		chtemp  = kelvin2Celsius(kch)
		maxtemp = kelvin2Celsius(kmax)
		mintemp = kelvin2Celsius(kmin)
		avgtemp = kelvin2Celsius(kavg)

		#print('k min/max/avg: ', kelvin2Celsius(kmin), kelvin2Celsius(kmax), kelvin2Celsius(kavg), kelvin2Celsius(kch))

		'''

		# Row major order
		hi   = thdata[96][128][0] # HI temp @ middle of screen crosshair
		lo   = thdata[96][128][1] # LO temp @ middle of screen crosshair
		chtemp = kelvin2Celsius(hi + (lo*256)) # Combine high and shifted low bytes


		#find the max temperature in the frame
		lomax  = thdata[...,1].max()
		posmax = thdata[...,1].argmax()
		#since argmax returns a linear index, convert back to row and col
		mcol,mrow = divmod(posmax,width)
		himax   = thdata[mcol][mrow][0]
		maxtemp = kelvin2Celsius(himax + (lomax*256)) # Combine high and shifted low bytes
		
		#find the lowest temperature in the frame
		lomin  = thdata[...,1].min()
		posmin = thdata[...,1].argmin()
		#since argmax returns a linear index, convert back to row and col
		lcol,lrow = divmod(posmin,width)
		himin   = thdata[lcol][lrow][0]
		mintemp = kelvin2Celsius(himin + (lomin*256))

		#print('ch, max, min: hi lo :', hi, lo, ',', himax, lomax, ",", himin, lomin)

		#find the average temperature in the frame
		loavg   = thdata[...,1].mean()
		hiavg   = thdata[...,0].mean()
		avgtemp = kelvin2Celsius(hiavg + (loavg*256)) # Combine high and shifted low bytes

		'''

		if chtemp < mintemp:
			print('    min error: ', mintemp, maxtemp, avgtemp, chtemp)

		if maxtemp < chtemp:
			print('    MAX error: ', mintemp, maxtemp, avgtemp, chtemp)

		# Convert the real image to RGB
		bgr = cv2.cvtColor(imdata,  cv2.COLOR_YUV2BGR_YUYV)
		#Contrast
		bgr = cv2.convertScaleAbs(bgr, alpha=alpha)#Contrast
		#bicubic interpolate, upscale and blur
		bgr = cv2.resize(bgr,(newWidth,newHeight),interpolation=cv2.INTER_CUBIC)#Scale up!
		if rad>0:
			bgr = cv2.blur(bgr,(rad,rad))


		#apply colormap - Optimize: remove unnecessary if checks from animation loop
		heatmap = cv2.applyColorMap(bgr, cmapCurrent)
		if colormap == 10:
			heatmap = cv2.cvtColor(heatmap, cv2.COLOR_BGR2RGB)


		#print(heatmap.shape)

		# draw crosshairs
		cv2.line(heatmap,(int(newWidth/2),int(newHeight/2)+20),\
		(int(newWidth/2),int(newHeight/2)-20),WHITE,2) #vline
		cv2.line(heatmap,(int(newWidth/2)+20,int(newHeight/2)),\
		(int(newWidth/2)-20,int(newHeight/2)),WHITE,2) #hline

		cv2.line(heatmap,(int(newWidth/2),int(newHeight/2)+20),\
		(int(newWidth/2),int(newHeight/2)-20),BLACK,1) #vline
		cv2.line(heatmap,(int(newWidth/2)+20,int(newHeight/2)),\
		(int(newWidth/2)-20,int(newHeight/2)),BLACK,1) #hline
		#show chtemp
		chtempCF = CorF(chtemp)
		cv2.putText(heatmap,str(chtempCF)+labelCF, (int(newWidth/2)+10, int(newHeight/2)-10),\
		font, 0.45, BLACK, 2, cv2.LINE_AA)
		cv2.putText(heatmap,str(chtempCF)+labelCF, (int(newWidth/2)+10, int(newHeight/2)-10),\
		font, 0.45, tempTextColor, 1, cv2.LINE_AA)

		frameCounter = frameCounter + 1

		if hud==True:
			# display black box for our data
			cv2.rectangle(heatmap, (0, 0),(166, 134), BLACK, -1)

			# put text in the box
			cv2.putText(heatmap,'Avg Temp: '+str(CorF(avgtemp))+labelCF, (10, 14),\
			font, 0.4, YELLOW, 1, cv2.LINE_AA)

			cv2.putText(heatmap,'Label Threshold: '+str(CorF(threshold))+labelCF, (10, 28),\
			font, 0.4, YELLOW, 1, cv2.LINE_AA)

			cv2.putText(heatmap,'Colormap: '+cmapText, (10, 42),\
			font, 0.4, YELLOW, 1, cv2.LINE_AA)

			cv2.putText(heatmap,'Blur: '+str(rad)+' ', (10, 56),\
			font, 0.4, YELLOW, 1, cv2.LINE_AA)

			cv2.putText(heatmap,'Scaling: '+str(scale)+' ', (10, 70),\
			font, 0.4, YELLOW, 1, cv2.LINE_AA)

			cv2.putText(heatmap,'Contrast: '+str(alpha)+' ', (10, 84),\
			font, 0.4, YELLOW, 1, cv2.LINE_AA)

			cv2.putText(heatmap,'Snapshot: '+snaptime+' ', (10, 98),\
			font, 0.4, YELLOW, 1, cv2.LINE_AA)

			cv2.putText(heatmap,'Recording: '+elapsed, (10, 112),\
			font, 0.4, rColor, 1, cv2.LINE_AA)

			nsB = time.time_ns()
			seconds = (nsB - nsA) / 1000000000
			fps = 'FPS: ' + str( round(frameCounter/seconds,1) )

			cv2.putText(heatmap, fps, (10, 128),\
			font, 0.4, YELLOW, 1, cv2.LINE_AA)

			if seconds > 60: # Reset counters every N seconds
				nsA = nsB
				frameCounter = 0

		
		#Yeah, this looks like we can probably do this next bit more efficiently!
		#display floating max temp (while keeping text visible @ display boundaries)
		if maxtemp > avgtemp+threshold:
			drawDotLabel(mrow, mcol, 'Max: '+str(CorF(maxtemp))+labelCF, RED)

		#display floating min temp (while keeping text visible @ display boundaries)
		if mintemp < avgtemp-threshold:
			drawDotLabel(lrow, lcol, 'Min: '+str(CorF(mintemp))+labelCF, BLUE)

		#display image
		cv2.imshow('Thermal',heatmap)

		if recording == True:
			elapsed = (time.time() - start)
			elapsed = time.strftime("%H:%M:%S", time.gmtime(elapsed)) 
			#print(elapsed)
			videoOut.write(heatmap)
		
		# Optimize key handling in anaimation loop
		# Make optimized pseudo case statement to reduce amount of unnecessary if checks every ms
		keyPress = cv2.waitKey(1)

		if keyPress <= -1: # check for most frequent return value (timeout) first 
			# print(keyPress)
			continue # keyPress timeout, don't do NN more if checks

		elif keyPress == ord('a'): #Increase blur radius
			rad += 1

		elif keyPress == ord('z'): #Decrease blur radius
			rad -= 1
			if rad < 0:
				rad = 0

		elif keyPress == ord('s'): #Increase threshold
			threshold += 1

		elif keyPress == ord('x'): #Decrease threashold
			threshold -= 1
			if threshold < 0:
				threshold = 0

		#print('threshold='+str(threshold))

		elif keyPress == ord('d'): #Increase scale
			reScale(1, True)

		elif keyPress == ord('c'): #Decrease scale
			reScale(-1, True)

		elif keyPress == ord('g'): # Toggle fullscreen
			toggleFullScreen()

		elif keyPress == ord('f'): #contrast+
			reAlpha(0.1)

		elif keyPress == ord('v'): #contrast-
			reAlpha(-0.1)

		elif keyPress == ord('h'): # Toggle HUD legend display on/off
			hud = not(hud)

		elif keyPress == ord('u'): # Toggle Celsius/Fahrenheit
			useCelsius = not(useCelsius)
			labelCF = ' C' if useCelsius==True else ' F'

		elif keyPress == ord('m'): #m to cycle forward through color maps
			reColormap(1)

		elif keyPress == ord('M'): #m to cycle backward through color maps
			reColormap(-1)

		elif keyPress == ord('r'): # Toggle recording start/stop
			recording = not(recording)
			if recording==True:
				videoOut = rec()
				start = time.time()
				rColor = RED # (40,40,250) # OFF_RED
			else:
				elapsed = "00:00:00"
				rColor = WHITE # (200,200,200) # OFF_WHITE

		elif keyPress == ord('p'): #p Take picture snapshot
			snaptime = snapshot(heatmap)

		elif keyPress == ord('q'):
			cap.release() # "capture" not defined
			cv2.destroyAllWindows()
			break  # Exit Python script 
		
